--
-- Smooth corner collisions in tiled top-down game
-- by Marcus 'ReFreezed' Thunström
--

local WORLD_WIDTH  = 8
local WORLD_HEIGHT = 8

local CAMERA_ZOOM = 50

local MOVE_SPEED = 5 -- Tiles per second.

local TAU = 2*math.pi

local function clamp(n, min, max)
	return math.min(math.max(n, min), max)
end

local function coordsToTileIndex(x, y)
	local tx = clamp(math.floor(x), 0, WORLD_WIDTH-1)
	local ty = clamp(math.floor(y), 0, WORLD_HEIGHT-1)
	local ti = ty*WORLD_WIDTH + tx + 1
	return ti
end

function love.load()
	player = {
		x      = .5*WORLD_WIDTH,
		y      = .3*WORLD_HEIGHT,
		radius = .3,
	}

	terrain = {
		1,1,1,1,1,1,1,1,
		1,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,1,
		1,0,0,0,0,0,0,1,
		1,0,0,1,0,0,0,1,
		1,0,0,1,1,0,0,1,
		1,0,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,
	}
	assert(#terrain == WORLD_WIDTH*WORLD_HEIGHT)

	local imageData = love.image.newImageData(WORLD_WIDTH, WORLD_HEIGHT)

	imageData:mapPixel(function(tx,ty, r,g,b,a)
		local value = terrain[coordsToTileIndex(tx, ty)]
		return value, value, value, 1
	end)

	love.graphics.setDefaultFilter("nearest", "nearest")
	terrainImage = love.graphics.newImage(imageData)
end

local function rotate(x,y, angle)
	local c = math.cos(angle)
	local s = math.sin(angle)
	return c*x-s*y, s*x+c*y
end

local function collideCircleWithWall(x,y, radius, wallNormalVectorAngle)
	-- The wall must to go through [0,0].

	-- Transform the coordinate system so that right (positive x) is outwards from the wall.
	x, y = rotate(x,y, -wallNormalVectorAngle)

	-- Move circle to right against the wall, plu a little extra.
	x = radius + .000001

	-- Transform the coordinate system back.
	x, y = rotate(x,y, wallNormalVectorAngle)

	return x, y
end

function love.update(dt)
	local LK = love.keyboard

	if LK.isDown("lctrl", "rctrl") then
		dt = dt / 10
	end

	-- Move the player.
	local left  = LK.isDown("left")
	local right = LK.isDown("right")
	local up    = LK.isDown("up")
	local down  = LK.isDown("down")

	local moveDirectionX = (left and -1 or 0) + (right and 1 or 0)
	local moveDirectionY = (up   and -1 or 0) + (down  and 1 or 0)

	if not (moveDirectionX == 0 and moveDirectionY == 0) then
		local angle = math.atan2(moveDirectionY, moveDirectionX)
		player.x    = player.x + MOVE_SPEED * math.cos(angle) * dt
		player.y    = player.y + MOVE_SPEED * math.sin(angle) * dt
	end

	--
	-- Handle slope collisions.
	--
	local radius              = player.radius
	collidedWithSlopeTriangle = nil

	local tx1 = math.floor(player.x-radius)
	local ty1 = math.floor(player.y-radius)
	local tx2 = math.ceil (player.x+radius-1)
	local ty2 = math.ceil (player.y+radius-1)

	for ty = ty1, ty2 do
		for tx = tx1, tx2 do
			local wallLeft  = terrain[coordsToTileIndex(tx-1, ty)] == 1
			local wallRight = terrain[coordsToTileIndex(tx+1, ty)] == 1
			local wallUp    = terrain[coordsToTileIndex(tx, ty-1)] == 1
			local wallDown  = terrain[coordsToTileIndex(tx, ty+1)] == 1

			local wallCount = (
				(wallLeft  and 1 or 0) +
				(wallRight and 1 or 0) +
				(wallUp    and 1 or 0) +
				(wallDown  and 1 or 0)
			)

			-- Only care about tiles with walls on two sides.
			if wallCount == 2 then
				local xInTile = player.x - tx
				local yInTile = player.y - ty

				local function handleSlopeCollision(wallSideX, wallSideY)
					local wallAngle = math.atan2(-wallSideY, -wallSideX) -- This should point out from the wall.
					local x, y      = collideCircleWithWall(xInTile-.5, yInTile-.5, radius, wallAngle)
					player.x        = tx + x + .5
					player.y        = ty + y + .5

					-- Show what slope we collided with.
					collidedWithSlopeTriangle = {tileCenterX=tx+.5, tileCenterY=ty+.5, wallAngle=wallAngle}
				end

				-- Get the distance, on one axis, from the center of the player to where the tangent
				-- would be if we were colliding with a slope. It's kind of a magical number for
				-- accounting for the player's radius in one of the checks.
				local diagonal = radius * math.sin(TAU/8)

				--                              Quick check what side of the         Limits so we're not too far off
				-- Which corner.                slope line we're on.                 the sides of the slope.
				-- |                            |                                    |
				if (wallLeft  and wallUp  ) and (xInTile+yInTile < 1+2*diagonal) and (xInTile > 0+diagonal and yInTile > 0+diagonal) then  handleSlopeCollision(-1,-1) end
				if (wallRight and wallUp  ) and (yInTile-xInTile < 0+2*diagonal) and (xInTile < 1-diagonal and yInTile > 0+diagonal) then  handleSlopeCollision( 1,-1) end
				if (wallRight and wallDown) and (xInTile+yInTile > 1-2*diagonal) and (xInTile < 1-diagonal and yInTile < 1-diagonal) then  handleSlopeCollision( 1, 1) end
				if (wallLeft  and wallDown) and (yInTile-xInTile > 0-2*diagonal) and (xInTile > 0+diagonal and yInTile < 1-diagonal) then  handleSlopeCollision(-1, 1) end
			end
		end
	end

	--
	-- Handle axis-aligned wall collisions.
	-- We do this after the slopes because these collisions are more authoritative.
	--
	local tx1 = math.floor(player.x-radius)
	local ty1 = math.floor(player.y-radius)
	local tx2 = math.ceil (player.x+radius-1)
	local ty2 = math.ceil (player.y+radius-1)

	for ty = ty1, ty2 do
		for tx = tx1, tx2 do
			if terrain[coordsToTileIndex(tx, ty)] == 1 then
				local xInTileClosestToPlayer = clamp(player.x, tx, tx+1)
				local yInTileClosestToPlayer = clamp(player.y, ty, ty+1)

				local dx = player.x - xInTileClosestToPlayer
				local dy = player.y - yInTileClosestToPlayer

				local distance = math.sqrt(dx*dx + dy*dy)

				if distance < radius then
					-- The player is intersecting the wall!
					local angle = math.atan2(dy, dx)
					local x, y  = collideCircleWithWall(dx,dy, radius, angle)
					player.x    = xInTileClosestToPlayer + x
					player.y    = yInTileClosestToPlayer + y
				end
			end
		end
	end
end

function love.draw()
	local LG                        = love.graphics
	local windowWidth, windowHeight = LG.getDimensions()

	--
	-- Draw world (centered on the screen).
	--
	LG.clear(.2, .2, .2)

	LG.translate(windowWidth/2, windowHeight/2)
	LG.scale(CAMERA_ZOOM)
	LG.translate(-WORLD_WIDTH/2, -WORLD_HEIGHT/2)

	LG.setColor(1, 1, 1)
	LG.draw(terrainImage)

	-- Show what slope we last collided with, if any.
	if collidedWithSlopeTriangle then
		LG.push()
		LG.translate(collidedWithSlopeTriangle.tileCenterX, collidedWithSlopeTriangle.tileCenterY)
		LG.rotate(collidedWithSlopeTriangle.wallAngle+TAU/8)
		LG.setColor(1, 0, 0)
		LG.polygon("fill", -.5,-.5, .5,.5, -.5,.5)
		LG.pop()
	end

	-- Draw player.
	LG.setColor(0, .5, 1)
	LG.circle("fill", player.x, player.y, player.radius, 64)

	--
	-- Draw HUD.
	--
	local text = string.format(
		"%.2f,%.2f\nPress ctrl for slow-mo",
		player.x, player.y
	)
	LG.origin()
	LG.setColor(1, 1, 1)
	LG.print(text, 5, 5)
end
